<?php

namespace Drupal\commerce_opp;

use Drupal\commerce_opp\Plugin\Commerce\PaymentType\OppSibsMultibanco;
use Drupal\commerce_opp\Transaction\Status\Pending;
use Drupal\commerce_opp\Transaction\Status\Rejected;
use Drupal\commerce_opp\Transaction\Status\SuccessOrPending;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\KeyValueStore\KeyValueFactory;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;

/**
 * Default Open Payment Platform service implementation.
 */
class OpenPaymentPlatformService implements OpenPaymentPlatformServiceInterface {

  /**
   * The payment storage.
   *
   * @var \Drupal\commerce_payment\PaymentStorageInterface
   */
  protected $paymentStorage;

  /**
   * The payment gateway storage.
   *
   * @var \Drupal\commerce_payment\PaymentGatewayStorageInterface
   */
  protected $paymentGatewayStorage;

  /**
   * The 'state' key/value store.
   *
   * @var \Drupal\Core\KeyValueStore\KeyValueStoreInterface
   */
  protected $keyValue;

  /**
   * The logger.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * Constructs a new OpenPaymentPlatformService object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\KeyValueStore\KeyValueFactory $key_value_factory
   *   The key/value store factory.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_channel_factory
   *   The logger channel factory.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, KeyValueFactory $key_value_factory, LoggerChannelFactoryInterface $logger_channel_factory) {
    $this->paymentStorage = $entity_type_manager->getStorage('commerce_payment');
    $this->paymentGatewayStorage = $entity_type_manager->getStorage('commerce_payment_gateway');
    $this->keyValue = $key_value_factory->get('state');
    $this->logger = $logger_channel_factory->get('commerce_opp');
  }

  /**
   * {@inheritdoc}
   */
  public function getOppGatewayIds($only_active = TRUE, $config_filter = [], $plugins = []) {
    $query = $this->paymentGatewayStorage->getQuery();
    $opp_plugin_ids = [
      'opp_copyandpay_bank',
      'opp_copyandpay_card',
      'opp_copyandpay_virtual',
      'opp_copyandpay_mbway',
      'opp_copyandpay_sibs_multibanco',
    ];
    $query->condition('plugin', $opp_plugin_ids, 'IN');
    if ($only_active) {
      $query->condition('status', TRUE);
    }
    foreach ($config_filter as $key => $value) {
      $query->condition('configuration.' . $key, $value);
    }
    if (!empty($plugins)) {
      $query->condition('plugin', $plugins, 'IN');
    }
    return $query->execute();
  }

  /**
   * {@inheritdoc}
   */
  public function deleteExpiredAuthorizations() {
    $opp_gateway_ids = $this->getOppGatewayIds(TRUE, [
      'cron_delete_expired_authorizations' => TRUE,
    ]);

    if (empty($opp_gateway_ids)) {
      return;
    }

    $query = $this->paymentStorage->getQuery();
    $query->condition('payment_gateway', $opp_gateway_ids, 'IN');
    $query->condition('state', 'authorization');
    $query->condition('expires', $this->keyValue->get('system.cron_last'), '<');
    $query->condition('type', 'opp_sibs_multibanco', '<>');
    $payments = $query->execute();
    if (empty($payments)) {
      return;
    }
    $payments = $this->paymentStorage->loadMultiple($payments);
    $this->paymentStorage->delete($payments);
  }

  /**
   * {@inheritdoc}
   */
  public function processPendingAuthorizations() {
    $opp_gateway_ids = $this->getOppGatewayIds(FALSE, [
      'cron_process_pending_authorizations' => TRUE,
    ]);

    if (empty($opp_gateway_ids)) {
      return;
    }

    $query = $this->paymentStorage->getQuery();
    $query->condition('payment_gateway', $opp_gateway_ids, 'IN');
    $query->condition('state', 'authorization');
    $query->condition('expires', $this->keyValue->get('system.cron_last'), '>=');
    $payments = $query->execute();
    if (empty($payments)) {
      return;
    }
    /** @var \Drupal\commerce_payment\Entity\PaymentInterface[] $payments */
    $payments = $this->paymentStorage->loadMultiple($payments);
    foreach ($payments as $payment) {
      /** @var \Drupal\commerce_opp\Plugin\Commerce\PaymentGateway\CopyAndPayInterface $gateway */
      $gateway = $payment->getPaymentGateway()->getPlugin();
      try {
        if ($payment->getType() instanceof OppSibsMultibanco) {
          /** @var \Drupal\commerce_opp\Plugin\Commerce\PaymentGateway\CopyAndPaySibsMultibanco $gateway */
          $gateway->checkAndUpdateTransactionStatus($payment);
          continue;
        }

        $payment_status = $gateway->getTransactionStatus($payment);
        if ($payment_status->isAsyncPayment() && $payment_status instanceof Pending) {
          $this->logger->info(sprintf('Skip processing of payment ID %s, as it is async type and in pending state.', $payment->id()));
          continue;
        }

        // The ID we receive in this response is different to the checkout ID.
        // The checkout ID was only a temporary remote value, in order to be
        // able to fetch the payment in this callback. Now, we have received the
        // real remote ID and will use it.
        $payment->setRemoteId($payment_status->getId());
        $payment->setRemoteState($payment_status->getCode());
        if ($payment_status instanceof SuccessOrPending) {
          $capture_transition = $payment->getState()->getWorkflow()->getTransition('capture');
          $payment->getState()->applyTransition($capture_transition);
          $payment->save();
        }
        elseif ($payment_status instanceof Rejected) {
          $void_transition = $payment->getState()->getWorkflow()->getTransition('void');
          $payment->getState()->applyTransition($void_transition);
          $payment->save();
        }
      }
      catch (\Exception $ex) {
        $this->logger->error($ex->getMessage());
      }
    }
  }

}
