<?php

namespace Drupal\commerce_opp\Plugin\Commerce\PaymentGateway;

use Drupal\commerce_opp\Transaction\Status\Factory;
use Drupal\commerce_opp\Transaction\Status\Rejected;
use Drupal\commerce_opp\Transaction\Status\SuccessOrPending;
use Drupal\commerce_opp\Transaction\Status\TransactionStatusInterface;
use Drupal\commerce_payment\Entity\PaymentInterface;
use Drupal\commerce_payment\Exception\InvalidResponseException;
use Drupal\commerce_payment\Exception\PaymentGatewayException;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\HasPaymentInstructionsInterface;
use Drupal\commerce_price\Price;
use Drupal\Core\Form\FormStateInterface;

/**
 * Provides a dedicated PAYFRAME gateway for SIBS MULTIBANCO virtual account.
 *
 * @CommercePaymentGateway(
 *   id = "opp_copyandpay_sibs_multibanco",
 *   label = "Open Payment Platform PAYFRAME: SIBS MULTIBANCO",
 *   display_label = "SIBS MULTIBANCO",
 *   forms = {
 *     "offsite-payment" = "Drupal\commerce_opp\PluginForm\CopyAndPayForm",
 *   },
 *   payment_type = "opp_sibs_multibanco",
 *   requires_billing_information = FALSE,
 * )
 */
class CopyAndPaySibsMultibanco extends CopyAndPayBase implements HasPaymentInstructionsInterface {

  /**
   * The date format used for sending date values to SIBS.
   *
   * This is currently used for these tow custom parameters:
   *   - SIBSMULTIBANCO_RefIntlDtTm
   *   - SIBSMULTIBANCO_RefLmtDtTm.
   *
   * @var string
   */
  const DATE_FORMAT = 'Y-m-d\TH:i:s.vP';

  /**
   * The date format used for parsing incoming date values from SIBS.
   *
   * The only reason, why this is different to the DATE_FORMAT constant defined
   * above, is the PHP restriction, that 'v' is an accepted format for
   * formatting date values as milliseconds, but unfortunately is not
   * recognized by \DateTime::createFromFormat(), which is very inconvenient and
   * will be hopefully fixed in future PHP releases.
   *
   * @var string
   */
  const DATE_FORMAT_PARSING = 'Y-m-d\TH:i:s.uP';

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'pa_expiration_period' => 3600,
      'sibs_payment_entity' => '',
      'test_mode' => 'EXTERNAL',
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form = parent::buildConfigurationForm($form, $form_state);

    $form['brands']['#access'] = FALSE;

    $form['pa_expiration_period'] = [
      '#type' => 'number',
      '#title' => $this->t('PA expiration time'),
      '#description' => $this->t('Expiration time of PA on SIBS side (in seconds).'),
      '#default_value' => $this->configuration['pa_expiration_period'],
    ];

    $form['sibs_payment_entity'] = [
      '#type' => 'textfield',
      '#title' => $this->t('SIBS Payment entity'),
      '#description' => $this->t('Payment entity to be provided by SIBS.'),
      '#default_value' => $this->configuration['sibs_payment_entity'],
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    parent::submitConfigurationForm($form, $form_state);

    if (!$form_state->getErrors()) {
      $values = $form_state->getValue($form['#parents']);
      $this->configuration['pa_expiration_period'] = $values['pa_expiration_period'];
      $this->configuration['sibs_payment_entity'] = $values['sibs_payment_entity'];
      $this->configuration['brands'] = ['SIBS_MULTIBANCO'];
    }
  }

  /**
   * {@inheritdoc}
   */
  public function prepareCheckout(array $params = [], PaymentInterface $payment = NULL) {
    if (!empty($this->configuration['sibs_payment_entity'])) {
      $params['customParameters[SIBSMULTIBANCO_PtmntEntty]'] = $this->configuration['sibs_payment_entity'];
    }

    if (!empty($this->configuration['pa_expiration_period'])) {
      $now = $this->time->getRequestTime();
      $expires = $now + $this->configuration['pa_expiration_period'];
      $params['customParameters[SIBSMULTIBANCO_RefIntlDtTm]'] = $this->dateFormatter->format($now, 'custom', self::DATE_FORMAT);
      $params['customParameters[SIBSMULTIBANCO_RefLmtDtTm]'] = $this->dateFormatter->format($expires, 'custom', self::DATE_FORMAT);
    }

    if ($this->getMode() == 'test') {
      $params['customParameters[SIBS_ENV]'] = 'QLY';
    }

    return parent::prepareCheckout($params, $payment);
  }

  /**
   * {@inheritdoc}
   */
  protected function alterTransactionStatus(TransactionStatusInterface $transaction_status, PaymentInterface $payment, array $json_response) {
    if ($transaction_status instanceof SuccessOrPending) {
      if (!empty($json_response['resultDetails']['pmtRef']) && $json_response['resultDetails']['AcquirerResponse'] == 'APPR') {
        $payment_ref_id = $json_response['resultDetails']['pmtRef'];
        if ($payment) {
          $payment->set('pmt_ref', $payment_ref_id);

          $request_time = $this->time->getRequestTime();
          $payment->setExpiresTime($request_time + $this->configuration['pa_expiration_period']);
          if (!empty($json_response['resultDetails']['refLmtDtTm'])) {
            $expireDateTime = \DateTime::createFromFormat(self::DATE_FORMAT_PARSING, $json_response['resultDetails']['refLmtDtTm']);
            if ($expireDateTime) {
              $payment->setExpiresTime($expireDateTime->getTimestamp());
            }
            else {
              $this->logger->warning($this->t('refLmtDtTm parameter of SIBS_MULTIBANCO transaction status seems not to be in @format date format: @refLmtDtTm (remote payment ID: @remote_id)', [
                '@format' => self::DATE_FORMAT_PARSING,
                '@refLmtDtTm' => $json_response['resultDetails']['refLmtDtTm'],
                '@remote_id' => $json_response['id'],
              ]));
            }
          }
          else {
            $this->logger->warning($this->t('SIBS_MULTIBANCO transaction status is missing the refLmtDtTm parameter (remote payment ID: @remote_id)', [
              '@remote_id' => $json_response['id'],
            ]));
          }
        }
      }
      else {
        throw new InvalidResponseException($this->t('No payment reference ID set for @checkout_id.', ['@checkout_id' => $payment->opp_checkout_id->value]));
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  protected function onReturnAction(PaymentInterface $payment, TransactionStatusInterface $payment_status) {
    $payment->setRemoteId($payment_status->getId());
    $payment->setRemoteState($payment_status->getCode());
    if ($payment_status instanceof SuccessOrPending) {
      $this->checkAndUpdateTransactionStatus($payment);
      $payment->save();
    }
    elseif ($payment_status instanceof Rejected) {
      $void_transition = $payment->getState()->getWorkflow()->getTransition('void');
      $payment->getState()->applyTransition($void_transition);
      $payment->save();
      throw new PaymentGatewayException($this->t('We could not complete your payment. Please try again or contact us if the problem persists.'));
    }
  }

  /**
   * Checks the transaction status and updates payment, if necessary.
   *
   * @param \Drupal\commerce_payment\Entity\PaymentInterface $payment
   *   The payment entity.
   */
  public function checkAndUpdateTransactionStatus(PaymentInterface $payment) {
    $transaction_report = $this->getTransactionReport($payment);
    foreach ($transaction_report['payments'] as $payment_info) {
      if ($payment_info['paymentType'] != 'RC') {
        continue;
      }
      $brand = $this->brandRepository->getBrand($payment_info['paymentBrand']);
      $payment_status = Factory::newInstance($payment_info['id'], $payment_info['result']['code'], $payment_info['result']['description'], $brand);
      if ($payment_status instanceof SuccessOrPending) {
        $payment->setAmount(new Price($payment_info['amount'], $payment_info['currency']));
        $capture_transition = $payment->getState()->getWorkflow()->getTransition('capture');
        $payment->getState()->applyTransition($capture_transition);
        $payment->save();
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function buildPaymentInstructions(PaymentInterface $payment) {
    $instructions = [];

    if (!$payment->get('pmt_ref')->isEmpty()) {
      $instructions = [
        '#theme' => 'sibs_multibanco_instructions',
        '#pmt_ref' => $payment->pmt_ref->value,
        '#amount' => $payment->getAmount(),
        '#validity_timestamp' => $payment->getExpiresTime(),
        '#sibs_payment_entity' => $this->configuration['sibs_payment_entity'],
      ];
    }

    return $instructions;
  }

  /**
   * {@inheritdoc}
   */
  protected function getBrandOptions() {
    $brand = $this->brandRepository->getBrand('SIBS_MULTIBANCO');
    return [$brand->getId() => $brand->getLabel()];
  }

}
